@@ -25,6 +25,18 @@ struct transaction_start_internal {
 	int status;
 };
 
+struct transaction_checkpoint_internal {
+	/* General state information. */
+	uint64_t machinenum;
+	int done;
+
+	/* Parameters used in creating a checkpoint. */
+	uint8_t seqnum[32];
+	uint8_t ckptnonce[32];
+	uint8_t whichkey;
+	int status;
+};
+
 struct transaction_commit_internal {
 	/* General state information. */
 	uint64_t machinenum;
@@ -42,6 +54,8 @@ static int storage_transaction_start(NETPACKET_CONNECTION *, uint64_t,
 static sendpacket_callback callback_getnonce_send;
 static handlepacket_callback callback_getnonce_response;
 static handlepacket_callback callback_start_response;
+static sendpacket_callback callback_checkpoint_send;
+static handlepacket_callback callback_checkpoint_response;
 static sendpacket_callback callback_commit_send;
 static handlepacket_callback callback_commit_response;
 
@@ -285,6 +299,131 @@ storage_transaction_start_fsck(NETPACKET_CONNECTION * NPC,
 	    NULL, seqnum, 2));
 }
 
+/**
+ * storage_transaction_checkpoint(machinenum, seqnum, ckptnonce, whichkey):
+ * Create a checkpoint ${ckptnonce} in the current write transaction, which
+ * has nonce ${seqnum}.  The value ${whichkey} is defined as in
+ * storage_transaction_commit.
+ */
+int
+storage_transaction_checkpoint(uint64_t machinenum, const uint8_t seqnum[32],
+    const uint8_t ckptnonce[32], uint8_t whichkey)
+{
+	struct transaction_checkpoint_internal C;
+	NETPACKET_CONNECTION * NPC;
+
+	/* Initialize transaction structure. */
+	C.machinenum = machinenum;
+	memcpy(C.seqnum, seqnum, 32);
+	memcpy(C.ckptnonce, ckptnonce, 32);
+	C.whichkey = whichkey;
+	C.done = 0;
+
+	/* Open netpacket connection. */
+	if ((NPC = netpacket_open()) == NULL)
+		goto err0;
+
+	/* Ask the netpacket layer to send a request and get a response. */
+	if (netpacket_op(NPC, callback_checkpoint_send, &C))
+		goto err1;
+
+	/* Wait until the checkpoint has been created (or failed). */
+	if (network_spin(&C.done))
+		goto err1;
+
+	/* Close netpacket connection. */
+	if (netpacket_close(NPC))
+		goto err0;
+
+	/* Report sequence number mismatch if necessary. */
+	switch (C.status) {
+	case 0:
+		/* Success. */
+		break;
+	case 1:
+		warn0("Sequence number mismatch creating checkpoint: "
+		    "Run --fsck");
+		goto err0;
+	default:
+		netproto_printerr(NETPROTO_STATUS_PROTERR);
+		goto err0;
+	}
+
+	/* Success! */
+	return (0);
+
+err1:
+	netpacket_close(NPC);
+err0:
+	/* Failure! */
+	return (-1);
+}
+
+static int
+callback_checkpoint_send(void * cookie, NETPACKET_CONNECTION * NPC)
+{
+	struct transaction_checkpoint_internal * C = cookie;
+
+	/* Ask the server to create a checkpoint. */
+	return (netpacket_transaction_checkpoint(NPC, C->machinenum,
+	    C->whichkey, C->ckptnonce, C->seqnum,
+	    callback_checkpoint_response));
+}
+
+static int
+callback_checkpoint_response(void * cookie, NETPACKET_CONNECTION * NPC,
+    int status, uint8_t packettype, const uint8_t * packetbuf,
+    size_t packetlen)
+{
+	struct transaction_checkpoint_internal * C = cookie;
+	int key;
+
+	(void)packetlen; /* UNUSED */
+	(void)NPC; /* UNUSED */
+
+	/* Handle read errors. */
+	if (status != NETWORK_STATUS_OK) {
+		netproto_printerr(status);
+		goto err0;
+	}
+
+	/* Make sure we received the right type of packet. */
+	if (packettype != NETPACKET_TRANSACTION_CHECKPOINT_RESPONSE)
+		goto err1;
+
+	/* Look up packet signing key. */
+	if ((key = key_lookup(C->whichkey)) == -1)
+		goto err0;
+
+	/* Verify packet hmac. */
+	switch (netpacket_hmac_verify(packettype, C->seqnum,
+	    packetbuf, 33, key)) {
+	case 1:
+		goto err1;
+	case -1:
+		goto err0;
+	}
+
+	/* Verify that this is a response to the right request. */
+	if (memcmp(&packetbuf[1], C->ckptnonce, 32))
+		goto err1;
+
+	/* Record status. */
+	C->status = packetbuf[0];
+
+	/* We're done! */
+	C->done = 1;
+
+	/* Success! */
+	return (0);
+
+err1:
+	netproto_printerr(NETPROTO_STATUS_PROTERR);
+err0:
+	/* Failure! */
+	return (-1);
+}
+
 /**
  * storage_transaction_commit(machinenum, seqnum, whichkey):
  * Commit the transaction ${seqnum} if it is the most recent uncommitted
@@ -336,7 +475,7 @@ callback_commit_send(void * cookie, NETPACKET_CONNECTION * NPC)
 {
 	struct transaction_commit_internal * C = cookie;
 
-	/* Ask the server if the file exists. */
+	/* Ask the server to commit the transaction. */
 	return (netpacket_transaction_commit(NPC, C->machinenum, C->whichkey,
 	    C->seqnum, callback_commit_response));
 }
